module Main where

type Bits = [Int]

swap :: (a,b) -> (b,a)
swap (a,b) = (b,a)

main :: IO ()
main = do
  raw <- getContents
  let ls = map toBits $ lines raw
    in do
      putStrLn $ "day3a: " ++ (show (solveA ls))
      putStrLn $ "day3b: " ++ (show (solveB ls))

toBits :: String -> Bits
toBits = map (\c -> if c == '0' then 0 else 1)

initTally :: Bits -> [(Int,Int)]
initTally s = take (length s) $ repeat (0,0)

solveA :: [Bits] -> Int
solveA [] = 0
solveA xs = finTally $ foldl incTally (initTally (head xs)) xs

myFlip :: Bits -> Bits
myFlip = map (\x -> if x == 0 then 1 else 0)

finTally :: [(Int,Int)] -> Int
finTally xs = gamma * delta
  where
    gamma = binToDec $ gs
    delta = binToDec $ myFlip gs
    gs = map f xs
    f (x,y) = if x > y then 0 else 1

binToDec :: [Int] -> Int
binToDec s = f 0 (reverse s)
  where
    f _ [] = 0
    f n (x:xs) = (x*(2^n)) + f (n+1) xs

incTally :: [(Int,Int)] -> Bits -> [(Int,Int)]
incTally xs ys = zipWith f xs ys
  where
    f (x,y) 0 = ((x+1),y)
    f (x,y) 1 = (x,(y+1))

solveB :: [Bits] -> Int
solveB xs = ogr * csr
  where
    ogr = binToDec $ finTally' (>) 0 1 xs
    csr = binToDec $ finTally' (<=) 0 0 xs

finTally' :: (Int -> Int -> Bool) -> Int -> Int -> [Bits] -> Bits
finTally' c n d [l] = l
finTally' c n d [] = error "empty list"
finTally' c n d ls = finTally' c (n+1) d $ filter (\b -> (b !! n) == k) ls
  where
    k = genKey c d $ map (drop n) ls

genKey :: (Int -> Int -> Bool) -> Int -> [Bits] -> Int
genKey c d xs = if x == y then d else (if c x y then 0 else 1)
  where
    (x,y) = foldl f (0,0) (map head xs)
    f (x,y) 0 = ((x+1),y)
    f (x,y) 1 = (x,(y+1))
