module Main where

data Space = Space Int (Int,Int) Bool
  deriving (Show, Eq)
type Board = [Space]
type Calls = [Int]

parseBoard :: [[Int]] -> Board
parseBoard ns = concat $ map f $ zip [0..] $ map (zip ([0..] :: [Int])) ns
  where f (i,ls) = map (\(j,v) -> Space v (i,j) False) ls

fixBoards :: String -> [[[Int]]]
fixBoards s = map fixBoard $ f ls
  where
    ls = dropWhile (== "") $ lines s
    f ls = if ls == [] then []
           else (:) (takeWhile (/= "") ls)
               $ f (dropWhile (== "") $ dropWhile (/= "") ls)

fixBoard :: [String] -> [[Int]]
fixBoard = map ((map read) . words)

parseCalls :: String -> [Int]
parseCalls s = map read $ words $ fixSep s

fixSep :: String -> String
fixSep = map (\c -> if c == ',' then ' ' else c)

main :: IO ()
main = do
  rawCalls <- getLine
  rawBoards <- getContents
  let calls = parseCalls rawCalls
      boards = map parseBoard $ fixBoards rawBoards
    in do
      putStrLn $ "day4a: " ++ show (performCalls calls boards)
      putStrLn $ "day4b: " ++ show (losersSum calls boards)

getState :: Space -> Bool
getState (Space _ _ b) = b

getRow :: Space -> Int
getRow (Space _ (x,_) _) = x

getCol :: Space -> Int
getCol (Space _ (_,y) _) = y

getVal :: Space -> Int
getVal (Space v _ _) = v

genRows :: [Space] -> [[Space]]
genRows ss = genLines getRow ss 0

genCols :: [Space] -> [[Space]]
genCols ss = genLines getCol ss 0

genLines :: (Space -> Int) -> [Space] -> Int -> [[Space]]
genLines f ss r = if ns == []
  then []
  else ns : (genLines f ss (r+1))
  where
    ns = filter (\s -> r == (f s)) ss

winningLine :: [Space] -> Bool
winningLine ss = all getState ss

hasWon :: Board -> Bool
hasWon b = any winningLine (genRows b ++ genCols b)

sumVals :: Board -> Int
sumVals b = sum $ map getVal $ filter (not . getState) b

markCall :: Int -> Board -> Board
markCall n b = map f b
  where
    f (Space v p s) = if v == n then (Space v p True) else (Space v p s)

performCalls :: [Int] -> [Board] -> Int
performCalls [] _ = error "empty calls"
performCalls (c:cs) bs = if wbs /= []
  then c * (sumVals (head wbs))
  else performCalls cs nbs
  where
    wbs = filter hasWon nbs
    nbs = map (markCall c) bs

losersSum :: [Int] -> [Board] -> Int
losersSum [] _ = error "empty calls"
losersSum (c:cs) bs = if length lbs == 1
  then performCalls cs lbs
  else losersSum cs nbs
  where
    lbs = filter (not . hasWon) nbs
    nbs = map (markCall c) bs
