module Main where

import Data.List
import Text.Parsec

type Point = (Int,Int)
data LinePoints =
    Straight [Point]
  | Diag [Point]

main :: IO ()
main = do
  raw <- getContents
  let input = validate $ map (parse lineP []) $ lines raw
  putStrLn $ "day5a: " ++ show (solve $ concatLP $ filter isStraight input)
  putStrLn $ "day5b: " ++ show (solve $ concatLP input)

concatLP :: [LinePoints] -> [Point]
concatLP [] = []
concatLP ((Straight ps):ls) = ps ++ (concatLP ls)
concatLP ((Diag ps):ls) = ps ++ (concatLP ls)

validate :: [Either ParseError a] -> [a]
validate [] = []
validate ((Left _):_) = error "invalid input"
validate ((Right b):bs) = b:(validate bs)

lineP :: Parsec String () LinePoints
lineP = do
  x1 <- many1 digit
  _ <- char ','
  y1 <- many1 digit
  _ <- string " -> "
  x2 <- many1 digit
  _ <- char ','
  y2 <- many1 digit
  return $ createSpots (read x1, read y1) (read x2, read y2)

range :: Int -> Int -> [Int]
range x y = [ i .. j ]
  where
    (i,j) = if x > y then (y,x) else (x,y)

createSpots :: Point -> Point -> LinePoints
createSpots (x1,y1) (x2,y2)
  | x1 == x2 = Straight [ (x1,j) | j <- range y1 y2]
  | y1 == y2 = Straight [ (i,y1) | i <- range x1 x2]
  | otherwise = Diag $ runDiag (x1,y1) (x2,y2)

runDiag :: Point -> Point -> [Point]
runDiag (x1,y1) (x2,y2)
  | y1 < y2 =
    if x1 < x2
    then take (x2-x1+1) $ iterate dl (x1,y1)
    else take (x1-x2+1) $ iterate dr (x1,y1)
  | otherwise =
    if x1 < x2
    then take (x2-x1+1) $ iterate dr (x2,y2)
    else take (x1-x2+1) $ iterate dl (x2,y2)
  where
    dl (x,y) = (x+1,y+1)
    dr (x,y) = (x-1,y+1)

rle :: Eq a => [a] -> [(Int,a)]
rle xs = rle' $ zip (repeat 1) xs

rle' :: Eq a => [(Int,a)] -> [(Int,a)]
rle' ((xv,xk):(yv,yk):xs) =
  if xk == yk
  then rle' $ ((xv+yv),xk) : xs
  else (xv,xk) : (rle' ((yv,yk):xs))
rle' xs = xs

isStraight :: LinePoints -> Bool
isStraight (Straight _) = True
isStraight _ = False

solve :: [Point] -> Int
solve ps = length $ filter (\x -> (fst x) >= 2) $ rle $ sort ps

