module Main where

import Data.List

data Result =
    Incomplete [Char]
  | Illegal Char
  | Complete

main :: IO ()
main = do
  raw <- getContents
  let input = map catLine $ lines raw
    in do
      putStrLn $ "day10a: " ++ (show $ solveA input)
      putStrLn $ "day10b: " ++ (show $ solveB input)

solveA :: [Result] -> Int
solveA ss = sum $ map score $ [ x | Illegal x <- ss ]

solveB :: [Result] -> Int
solveB ss = getMid $ sort $ map acScore [ x | Incomplete x <- ss ]

acScore :: [Char] -> Int
acScore = acScore' 0

acScore' :: Int -> [Char] -> Int
acScore' n [] = n
acScore' n (c:cs) = acScore' (x+(5*n)) cs
  where
    x = case c of
      '(' -> 1
      '[' -> 2
      '{' -> 3
      '<' -> 4
      _ -> 0

catLine :: String -> Result
catLine = catLine' []

catLine' :: String -> String -> Result
catLine' [] [] = Complete
catLine' xs [] = Incomplete xs
catLine' xs (s:ss) =
  case isOpen s of
    True -> catLine' (s:xs) ss
    False -> if xs == []
      then error "close with no matching open"
      else let x = head xs
        in if match x s
          then catLine' (tail xs) ss
          else Illegal s

getMid :: [a] -> a
getMid [] = error "empty list"
getMid xs = xs !! n
  where
    n = (length xs) `div` 2

score :: Char -> Int
score ')' = 3
score ']' = 57
score '}' = 1197
score '>' = 25137
score _ = 0

match :: Char -> Char -> Bool
match x y = ( (x,y) `elem` ls || (y,x) `elem` ls)
  where
    ls = [
        ('(',')')
      , ('{','}')
      , ('<','>')
      , ('[',']')
      ]

isOpen :: Char -> Bool
isOpen = flip elem $ ['(','{','<','[']

