module Main where

import Data.List
import qualified Data.Map as M

type Db = M.Map (Int,Int) Oct

data Oct = Oct {
    value :: Int
  , total :: Int
  , flash :: Bool
  , nbs :: [(Int,Int)]
  } deriving Show

main :: IO ()
main = do
  raw <- getContents
  let input = map (map (read . (:[]))) $ lines raw
      vl = mkValList input
    in do
      putStrLn $ "day11a: " ++ (show $ solveA vl)
      putStrLn $ "day11b: " ++ (show $ solveB vl)

solveA :: Db -> Int
solveA db = sum $ map total $ M.elems $ (iterate step db) !! 100

solveB :: Db -> Int
solveB = go 0
  where
    go n db = if allFlash db then n else go (n+1) $ step db

allFlash :: Db -> Bool
allFlash db = all flash $ map snd $ M.toList db

step :: Db -> Db
step db = bfs incOct inits ndb
  where
    inits = M.keys ndb
    ndb = M.map unflash db

bfs :: (Oct -> Oct) -> [(Int,Int)] -> Db -> Db
bfs _ [] db = db
bfs f (k:ks) db = bfs f (ks++nks) ndb
  where
    tot = total (db M.! k)
    ndb = M.adjust f k db
    nme = ndb M.! k
    nks = if (flash nme) && not (flash (db M.! k))
          then nbs nme
          else []

incOct :: Oct -> Oct
incOct o = o
  { value = (value o) + 1
  , flash = (value o) >= 9
  , total = (total o) +
      if (value o) >= 9
         && not (flash o)
      then 1
      else 0
  }

unflash :: Oct -> Oct
unflash o = o
  { flash = False
  , value = if (flash o) then 0 else value o
  }

-- io and parsing

mkValList :: [[Int]] -> Db
mkValList [] = M.empty
mkValList raw =
  (M.fromList . concat)
  $ map (\(y, vs) ->
    map (\(x, v) ->
      ((x,y), (Oct v 0 False (mkDeltas w h (x,y)))))
    $ zip [0.. ] vs)
  $ zip [0..] $ raw
  where
    w = length (head raw)
    h = length raw

mkDeltas :: Int -> Int -> (Int,Int) -> [(Int,Int)]
mkDeltas w h (x,y) = filter
  (\(i,j) -> (i >= 0 && i < w && j >= 0 && j < h))
  [ (i,j) | i <- [(x-1) .. (x+1)]
          , j <- [(y-1) .. (y+1)]
          , (i /= x || j /= y)]

