module Main where

import Data.Char
import Data.List
import qualified Data.Map as M

type Db = M.Map String Node
data Path =
    Simple [String]
  | Extra [String]
    deriving Show
data Node = Node {
    big :: Bool
  , nbs :: [String]
  } deriving Show

main :: IO ()
main = do
  raw <- getContents
  let input = findPaths [ Simple ["start"] ] [] $ mkDb $ map splitInp $ lines raw
    in do
      putStrLn $ "day12a: " ++ (show $ solveA input)
      putStrLn $ "day12b: " ++ (show $ solveB input)

solveA :: [Path] -> Int
solveA ps = length $ [ x | Simple x <- ps ]

solveB :: [Path] -> Int
solveB = length

findPaths :: [Path] -> [Path] -> Db -> [Path]
findPaths [] comp _ = comp
findPaths (k:ks) comp db =
  if rec k == "end"
  then findPaths ks (k:comp) db
  else case k of
    (Extra ps) -> go stepExtra ps
    (Simple ps) -> go stepSimple ps
  where
    rec (Extra (k:_)) = k
    rec (Simple (k:_)) = k
    rec _ = error "empty path"
    go f ps =  findPaths ((f db ps) ++ ks) comp db

stepExtra :: Db -> [String] -> [Path]
stepExtra _ [] = error "empty list"
stepExtra db ps = map (Extra . (flip (:) ps)) (l (head ps) db)
  where
    isBig s = isUpper $ head s
    l k db = filter (\n -> (n /= "start") && ((isBig n) || not (n `elem` ps))) $ nbs $ db M.! k

stepSimple :: Db -> [String] -> [Path]
stepSimple _ [] = error "empty list"
stepSimple db ps = map (\p -> (c (p:ps))) (l (head ps) db)
  where
    isBig s = isUpper $ head s
    c ps = if isBig (head ps)
      then Simple ps
      else
        let smalls = filter (not . isBig) ps
        in if any (== (head smalls)) (tail smalls)
           then Extra ps
           else Simple ps
    l k db = filter (/= "start") $ nbs $ db M.! k

-- io and parsing

splitInp :: String -> (String,String)
splitInp s = (\(a,b) -> (a, drop 1 b)) $ splitAt n s
  where
    n = findIn '-' s

findIn :: Char -> String -> Int
findIn c s = go 0 c s
  where
    go n c (s:ss) = if c == s then n else go (n+1) c ss
    go n c [] = error "bad parse"

mkDb :: [(String,String)] -> Db
mkDb adjs = foldl f (M.fromList $ (map mkInitDb $ getUniques adjs)) adjs
  where
    f db (a,b) =
      M.adjust (\n -> n { nbs = (b:(nbs n)) }) a
      $ M.adjust (\n -> n { nbs = (a:(nbs n)) }) b
      $ db

mkInitDb :: String -> (String,Node)
mkInitDb s = (,) s
  $ Node (isUpper (head s)) []

getUniques :: [(String,String)] -> [String]
getUniques is = nub $ foldl (\t (a,b) -> (a:b:t)) [] is
