module Main where

import Data.List
import System.IO

data Fold =
    Hori Int
  | Vert Int

main :: IO ()
main = do
  raw <- getContents
  let dots = parseDots $ takeWhile (/= "") (lines raw)
      folds = parseFolds $ tail $ (dropWhile (/= "")) (lines raw)
  putStrLn $ "day13a: " ++ (show $ solveA dots (head folds))
  putStrLn $ "day13b:"
  mapM_ putStrLn $ solveB dots folds

solveA :: [(Int,Int)] -> Fold -> Int
solveA dots f = length $ nub $ map (crease f) dots

solveB :: [(Int,Int)] -> [Fold] -> [String]
solveB dots folds = showPaper $ (foldl (\d f -> map (crease f) d)) dots folds

crease :: Fold -> (Int,Int) -> (Int,Int)
crease (Hori h) (x,y) =
  if h >= x
  then (x,y)
  else ((h-(x-h)), y)
crease (Vert v) (x,y) =
  if v >= y
  then (x,y)
  else (x, (v-(y-v)))

parseFolds :: [String] -> [Fold]
parseFolds = map f
  where
    g (d:'=':t) = (if d == 'y' then Vert else Hori) (read t)
    f = g . (drop 11)

parseDots :: [String] -> [(Int,Int)]
parseDots = map f
  where
    g (x:y:_) = ((read x), (read y))
    f s = g $ words $ map (\c -> if c == ',' then ' ' else c) s

showPaper :: [(Int,Int)] -> [String]
showPaper db = map (\y -> mkRow y) [0..mH]
  where
    mW = maximum $ map fst $ db
    mH = maximum $ map snd $ db
    f c = if c then '▉' else ' '
    mkRow y = [ f $ (x,y) `elem` db | x <- [0..mW] ]
