module Main where

import Data.List
import qualified Data.Map as M
import Data.Map.Strict (insertWith)

type Rules = M.Map (Char,Char) Char
type Polymer = M.Map (Char,Char) Integer
type Freqs = M.Map Char Integer

main :: IO ()
main = do
  raw <- getContents
  let input = parseMatches $ drop 2 $ lines raw
      init = parsePolymer $ head $ lines raw
  putStrLn $ "day14a: " ++ (show $ solveA input init)
  putStrLn $ "day14b: " ++ (show $ solveB input init)

solveA :: Rules -> Polymer -> Integer
solveA = solve 10

solveB :: Rules -> Polymer -> Integer
solveB = solve 40

solve :: Int -> Rules -> Polymer -> Integer
solve n rs p = (maximum nums) - (minimum nums)
  where
    nums = M.elems $ getFreqs $ (iterate (step rs) p) !! n

step :: Rules -> Polymer -> Polymer
step rs p = foldr ($) M.empty $ map f $ M.toList p
  where
    f ((a,b),i) = let c = rs M.! (a,b)
                   in (insertWith (+) (a,c) i) . (insertWith (+) (c,b) i)

getFreqs :: Polymer -> Freqs
getFreqs p = M.map (\x -> (x+1) `div` 2) $ foldr ($) M.empty $ map f $ M.toList p
  where
    f ((a,b),i) = (insertWith (+) a i) . (insertWith (+) b i)

parsePolymer :: String -> Polymer
parsePolymer s = foldr f M.empty $ pairs s
  where
    pairs s = zip s (tail s)
    f p = M.insertWith (+) p 1

parseMatches :: [String] -> Rules
parseMatches ss = M.fromList $ map f ss
  where
    f s = ((s !! 0, s !! 1), s !! 6)
