module Main where

-- to do - do a more efficient djikstras impl

import qualified Data.PSQueue as P
import Debug.Trace
import Data.List
import qualified Data.Map as M
import Data.Map.Strict (insertWith)

type Point = (Int,Int)
type Graph = M.Map Point Vertex
type Queue = M.Map Point QueueNode
type PQ = P.PSQ Point Int

data QueueNode = QueueNode
  { qDist :: Int
  , qParent :: Maybe Point
  , qProc :: Bool
  } deriving Show

data Vertex = Vertex
  { vKey :: Point
  , vWeight :: Int
  , vAdjs :: [Point]
  } deriving Show

main :: IO ()
main = do
  raw <- getContents
  let input = map (map (read . (:[]))) $ lines raw
      vl = mkValList input
      vl2 = mkValList $ extendList input
      end s = (,) (((length (head input))*s) - 1) (((length input)*s) - 1)
  putStrLn $ "day15a: " ++ (show $ solveA vl $ (end 1))
  putStrLn $ "day15b: " ++ (show $ solveB vl2 $ (end 5))

solveA :: Graph -> Point -> Int
solveA g end = findWeight g end

solveB :: Graph -> Point -> Int
solveB = solveA

showPath :: Queue -> Point -> [Point]
showPath q p
  | p == (0,0) = [(0,0)]
  | otherwise = case qParent (q M.! p) of
      Nothing -> []
      Just k -> (p:(showPath q k))

findWeight :: Graph -> Point -> Int
findWeight g end = tracePath end q
  where
    (_,q) = dijk (0,0) end initQueue g (P.singleton (0,0) 0)

initQueue :: Queue
initQueue = M.singleton (0,0) $ QueueNode 0 Nothing False

tracePath :: Point -> Queue -> Int
tracePath c q = qDist (q M.! c)

getMin :: PQ -> (Point,PQ)
getMin pq = case P.minView pq of
  Nothing -> error "empty pq"
  Just ((k P.:-> _), pq') -> (k,pq')

dijk :: Point -> Point -> Queue -> Graph -> PQ -> (Point,Queue)
dijk start end q g pq
  | P.null pq = error "no path found"
  | otherwise =
      let (k,pq') = getMin pq
      in if qProc (q M.! k)
      then dijk start end q g pq'
      else if k == end
        then (k,q)
        else let
            as = vAdjs (g M.! k)
            q' = updateQ k q g
          in dijk start end q' g (addAdjs q' as pq')

updateQ :: Point -> Queue -> Graph -> Queue
updateQ p q g = foldr ($) q $ (h:(map f as))
  where
    h = M.adjust (\o -> o { qProc = True }) p
    as = vAdjs (g M.! p)
    f a = let l = qDist (q M.! p)
              d = l + (vWeight $ g M.! a)
      in if betterRoute d a q
      then insertWith
        (\o n -> n { qProc = qProc o })
        a
        (QueueNode { qParent = Just p, qDist = d, qProc = False})
      else id

betterRoute :: Int -> Point -> Queue -> Bool
betterRoute d k q =
  if not (M.member k q)
  then True
  else d < (qDist (q M.! k))

addAdjs :: Queue -> [Point] -> PQ -> PQ
addAdjs _ [] pq = pq
addAdjs q (a:as) pq = addAdjs q as (P.insert a (qDist (q M.! a)) pq)

-- io stuff

modWeight :: Int -> Int
modWeight n = if n >= 9 then modWeight (n-9) else n+1

extendList :: [[Int]] -> [[Int]]
extendList rs = concat $ take 5 $ iterate (map (map modWeight)) $ map extend rs

extend :: [Int] -> [Int]
extend cs = concat $ take 5 $ iterate (map modWeight) cs

mkValList :: [[Int]] -> Graph
mkValList [] = M.empty
mkValList raw =
  (M.fromList . concat)
  $ map (\(y, vs) ->
    map (\(x, v) ->
      ((x,y), (Vertex (x,y) v (mkDeltas w h (x,y)))))
    $ zip [0.. ] vs)
  $ zip [0..] $ raw
  where
    w = length (head raw)
    h = length raw

mkDeltas :: Int -> Int -> (Int,Int) -> [(Int,Int)]
mkDeltas w h (x,y) = filter
  (\(i,j) -> (i >= 0 && i < w && j >= 0 && j < h))
  [ (i,j) | i <- [(x-1) .. (x+1)]
          , j <- [(y-1) .. (y+1)]
          , (i == x || j == y)]
